function [sp,p,mu,out]= NewDORE(H,Ht,y,r,varargin)

% Double Overrelaxation ECME (DORE) Thresholding routine
% Coded by Kun Qiu (kqiu@iastate.edu) initially and then revised by 
% Renliang Gu (renliang@iastate.edu)
% updated Jul 6, 2012
% 
% Function usage
% =======================================================
% INPUT(compulsory):
% H:    the sensing matrix (H can be either an matrix or function handle 
%       that computes H*x)
% Ht:   the adjoint function handle that computes H^T*y (if H is a matrix, 
%       just input [])
% y:    the measurement column vector
% r:    the sparsity level of the signal to be recovered
% 
% INPUT(optional):
% 'IsHOrthonormal': whether H is orthonormal: 
%                       0: no
%                       1: yes(default)
% 'InitialSig':     the mx1 initial signal estimate(default=zeros(m,1))
% 'Thresh':         tolerance threshold for stopping the algorithm
%                       (default=1e-14)
% 'Visibility':     Option to view the reconstruction process 
%                       (valued in {0,1})
%                       0: do not view(default)
%                       1: do view
% ========================================================
% OUTPUT:
% s_hat:            the signal estimate
% A_index:          the estimated support set
% Count:            Count of the number of EM iterations
% loglikelihood:    the value of the log-likelihood function of the final 
%                   estimate
% delta2:           the variance component estimate
% ========================================================

if (nargin-length(varargin))~=4
    error('Missing required inputs!');
end

while isa(varargin{1},'cell')
    varargin=varargin{1};
end

if ~isa(H, 'function_handle')
    H_temp=H; Ht=@(x) H'*x; H=@(x) H*x;
else H_temp=[]; end

N=length(y);
m=length(Ht(y));
normy=norm(y);

%Setting default values for the optional inputs
sInit=zeros(m,1);
thresh=1e-14;
Visibility=0;
maxitr=10000;
str='';
adjustLim=160;
shrink=0.9;

%Read the optional inputs
if (rem(length(varargin),2)==1)
    error('Optional inputs must go by pairs!');
else
    for i=1:2:(length(varargin)-1)
        switch upper(varargin{i})
            case upper('StepMode')
                % abc
                % a: double mu at first step
                % b: shrink mu during iteration 
                % c: mu initiated by 1, 2 to specify user step
                stepMode=varargin{i+1};
            case upper('InitialSig')
                sInit=varargin{i+1};
            case upper('Thresh')
                thresh=varargin{i+1};
            case upper('Visibility')
                Visibility=varargin{i+1};
            case upper('MaxItr')
                maxitr=varargin{i+1};
            case upper('step')
                step=varargin{i+1};
            otherwise
                error(['Unrecognized optional input: ''' varargin{i} '''']);
        end
    end
end

adjustNum=0;
adjust=zeros(adjustLim,1);
resVec=zeros(maxitr,1);
likelihood=zeros(maxitr,1);
gap=zeros(maxitr,1);
DORisGood=zeros(maxitr,1);
time=zeros(maxitr,1);
ECMEfail=zeros(maxitr,1);
stepModec=mod(stepMode,10);
stepMode=floor(stepMode/10);
stepModeb=mod(stepMode,10);
stepModea=floor(stepMode/10);
fprintf('stepMode: %d-%d-%d\nr=%d\n',stepModea,stepModeb,stepModec,r);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch stepModec
    case 2
        mu=step;
    case 1
        mu=1;
    case 0
        if(~isempty(H_temp))
            mu=inv(H_temp*H_temp');
        else
            ee=zeros(N,1);
            ind=randperm(N,10);
            rho=expectHHt(H,Ht,N,m,'H',50);
            fprintf('rho=%g\n',rho);
            mu=1/rho;
        end
    otherwise
        error('stepModec=%d is not implemented yet!\n',stepModec);
end
clear H_temp;

tic;
%Initialization
sp1=sInit;
[~,idx]=sort(abs(sp1),'descend');
sp1(idx(r+1:end))=0;
Hsp1=H(sp1);

sp=sp1+Ht(mu*(y-Hsp1));
[~,idx]=sort(abs(sp),'descend');
sp(idx(r+1:end))=0;
Hsp=H(sp);
sqrErrp=(y-Hsp)'*(y-Hsp);
%   loglikelihoodp=-N/2*log(2*pi*deltap)-N/2;

exit_flag=0;
p=0; rec=0;
needDouble=true;

while ~exit_flag
    p=p+1;
    %STEP 1
    while(1)
        sHat=sp+Ht(mu*(y-Hsp));
        [~,idx]=sort(abs(sHat),'descend');
        sHat(idx(r+1:end))=0;
        HsHat=H(sHat);
        sqrErrHat=(y-HsHat)'*(y-HsHat);

        if(sqrErrHat<=sqrErrp)
            if(needDouble &&(stepModea==1)&&(length(mu(:))==1))
                mu=mu*2;
                fprintf('mu is doubled to %g\n',mu);
                str='';
                adjustNum=adjustNum+1;
                adjust(adjustNum)=p;
            else
                break;
            end
        else
            needDouble=false;
            if((stepModeb==1)&&(length(mu(:))==1))
                mu=mu*shrink;
                fprintf('mu is shrinked to %g\n',mu);
                str='';
                adjustNum=adjustNum+1;
                adjust(adjustNum)=p;
                if(adjustNum==adjustLim)
                    fprintf('Too many time adjust, Error!\n');
                    str='';
                    break;
                end
            else
                break;
            end
        end
    end
    if(sqrErrHat>sqrErrp) ECMEfail(p)=1; end

    %STEP 2
    alpha1=(HsHat-Hsp)'*mu*(y-HsHat)/((HsHat-Hsp)'*mu*(HsHat-Hsp));
    zBar=sHat+alpha1*(sHat-sp);
    HzBar=HsHat+alpha1*(HsHat-Hsp);

    %STEP 3
    alpha2=(HzBar-Hsp1)'*mu*(y-HzBar)/((HzBar-Hsp1)'*mu*(HzBar-Hsp1));
    zBar=zBar+alpha2*(zBar-sp1);

    %STEP 4
    sTilde=zBar;
    [~,idx]=sort(abs(sTilde),'descend');
    sTilde(idx(r+1:end))=0;
    HsTilde=H(sTilde);
    sqrErrTilde=(y-HsTilde)'*(y-HsTilde);

    sp1=sp; Hsp1=Hsp;

    %STEP 5
    if sqrErrTilde>=sqrErrHat
        sp=sHat;
        Hsp=HsHat;
        sqrErrp=sqrErrHat;
    else
        sp=sTilde;
        Hsp=HsTilde;
        sqrErrp=sqrErrTilde;
        DORisGood(p)=1;
    end

    resVec(p)=sqrt(sqrErrp)/normy;
    likelihood(p)=-N/2*log(sqrErrp/N);
    %gap(p)=norm(sp1-sp)^2/m;
    if(p>=2) gap(p)=abs(likelihood(p)-likelihood(p-1));
    else gap(p)=likelihood(1); end
    if ((Visibility) && (abs(toc-rec)>3))
        rec=toc;
        for temp=1:length(str); fprintf('\b'); end
        str=sprintf('Itr=%d, mu=%g, gap=%g, res=%f   ',...
            p,mu,gap(p),resVec(p));
        fprintf('%s',str);
    end
    if (gap(p)<thresh || p>=maxitr); exit_flag=1; end
    time(p)=toc;
end

fprintf('\n\n');
out.resVec=resVec(1:p);
out.DOREisGood=DORisGood(1:p);
out.time=time(1:p);
out.ECMEfail=ECMEfail(1:p);
out.gap=gap(1:p);
out.adjust=adjust(1:adjustNum);
out.likelihood=likelihood(1:p);

